#include <GNSS.h>
#include <SD.h>

//Define CONTINUOUS_WRITE 0 to spreat file when lose fix.
#define CONTINUOUS_WRITE        0

static String dataFolder("/LogData");  //Default NMEA Log folder name
static String logFile;                 //NMEA file full path name
static int ledPin = 0;                 //NavSpark build-in LED 
static bool fileCreated = false;

//variables to count NMEA interval
struct NmeaInterval
{
  uint8_t gga; 
  uint8_t gll; 
  uint8_t gsa; 
  uint8_t gsv; 
  uint8_t rmc; 
  uint8_t vtg; 
  uint8_t zda; 
};
NmeaInterval nmeaInterval = { 0, 0, 0, 0, 0, 0, 0 };
NmeaInterval nmeacount = { 0, 0, 0, 0, 0, 0, 0 };

//log status to control LED toggle.
enum LogStatus
{
  NotFixed,
  Fixed,
  Logging
};
LogStatus logStatus = NotFixed;

void setup() {
  //Initialize GNSS device parameter
  GnssConf.setNavMode(STGNSS_NAV_MODE_AUTO);
  GnssConf.setUpdateRate(STGNSS_POSITION_UPDATE_RATE_1HZ);
  GnssConf.setDopMaskMode(STGNSS_DOP_MASK_AUTO);
  GnssConf.setPdopMask(30.0);
  GnssConf.setHdopMask(30.0);
  GnssConf.setGdopMask(30.0); 
  GnssConf.setMessageInterval(1, 0, 1, 1, 1, 0, 0); //Set NMEA gga, gsa, gsv, rmc output every 1 second.
  GnssConf.init();

  //Initial Serial for console
  Serial.begin(115200);
  Serial.println(" ");
  Serial.println("========== System Start ==========");

  //Initial LED
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, HIGH);
  
  //Initial SD Class for SD Card using
  SD.begin();
  bool b = SD.mkdir(dataFolder.c_str());
  Serial.print("Create Folder Status : ");
  Serial.println(b);
}

//Timer function for toggle build-in LED
void timer0_func(void) {
  static bool ledToggle = true;
  digitalWrite(ledPin, (ledToggle) ? HIGH : LOW);
  ledToggle = (ledToggle) ? false : true;
}

void loop() {
  // put your main code here, to run repeatedly:
  static LogStatus lsatLogStatus = NotFixed;
  if(lsatLogStatus == logStatus)
  {
    return;
  }

  //If log status change, setup timer for toggle LED.
  switch(logStatus)
  {
    case NotFixed:
      Timer0.stop();
      digitalWrite(ledPin, HIGH);
      break;
    case Fixed:
      Timer0.every(1000, timer0_func); // setup timer 0 for toggle led.
      break;
    case Logging:
      Timer0.every(100, timer0_func); // setup timer 0 for toggle led.
      break;
  }
  lsatLogStatus = logStatus;
}

int GenerateNMEA(String& logContent) {
  char buf[128] = {0};
  int length = 0;

  if(nmeaInterval.gga != 0)
  {
    if(nmeacount.gga <= 0 ||  nmeacount.gga > nmeaInterval.gga)
    {
      buf[0] = 0;
      length += gv8_get_gga(buf);
      logContent += buf;
      nmeacount.gga = nmeaInterval.gga;
    }
    --nmeacount.gga;
  } 
  
  if(nmeaInterval.gsa != 0)
  {
    if(nmeacount.gsa <= 0 ||  nmeacount.gsa > nmeaInterval.gsa)
    {
      buf[0] = 0;
      length += gv8_get_GPS_gsa(buf);
      logContent += buf;
 #if defined(ST_CONST_SEL) && (ST_CONST_SEL==3)     
      length += gv8_get_GLONASS_gsa(buf);
      logContent += buf;
#endif
#if defined(ST_CONST_SEL) && (ST_CONST_SEL==2 || ST_CONST_SEL==4)
      length += gv8_get_BD_gsa(buf);
      logContent += buf;
#endif
      nmeacount.gsa = nmeaInterval.gsa;
    }
    --nmeacount.gsa;
  } 

  if(nmeaInterval.gsv != 0)
  {
    if(nmeacount.gsv <= 0 ||  nmeacount.gsv > nmeaInterval.gsv)
    {
      int returnLen = 0;
      buf[0] = 0;
      returnLen = gv8_get_GPS_gsv(1, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }
      returnLen = gv8_get_GPS_gsv(2, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }
      returnLen = gv8_get_GPS_gsv(3, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }
      returnLen = gv8_get_GPS_gsv(4, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }      
#if defined(ST_CONST_SEL) && (ST_CONST_SEL==3)
      returnLen = gv8_get_GLONASS_gsv(1, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }   
      returnLen = gv8_get_GLONASS_gsv(2, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }   
      returnLen = gv8_get_GLONASS_gsv(3, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }   
      returnLen = gv8_get_GLONASS_gsv(4, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }   
#endif
#if defined(ST_CONST_SEL) && (ST_CONST_SEL==2 || ST_CONST_SEL==4)
      returnLen = gv8_get_BD_gsv(1, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      } 
      returnLen = gv8_get_BD_gsv(2, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }       
      returnLen = gv8_get_BD_gsv(3, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      }       
      returnLen = gv8_get_BD_gsv(4, buf);
      if(returnLen) {
        length += returnLen;
        logContent += buf;
      } 
#endif                 
      nmeacount.gsv = nmeaInterval.gsv;
    }
    --nmeacount.gsv;
  }
 
  if(nmeaInterval.rmc != 0)
  {
    if(nmeacount.rmc <= 0 ||  nmeacount.rmc > nmeaInterval.rmc)
    {
      buf[0] = 0;
      length += gv8_get_rmc(buf);
      logContent += buf;
      nmeacount.rmc = nmeaInterval.rmc;
    }
    --nmeacount.rmc;
  } 
  return length;
}

void CreateLogFile() {
  char pattern[32];
  //Using YearMonthDay-HourMinuteSecond.txt for logging name.
  sprintf(pattern, "%04d%02d%02d-%02d%02d%02d.txt", GnssInfo.date.year(), GnssInfo.date.month(), GnssInfo.date.day(), 
    GnssInfo.time.hour(), GnssInfo.time.minute(), GnssInfo.time.second());
  
  //Show log file path name in console
  logFile = dataFolder + "/";
  logFile += pattern;
  Serial.println(logFile);
  
  //Open log file to check SD card is ready
  File file = SD.open(logFile.c_str(), FILE_WRITE);
  if(file)
  { 
    file.close(); 
    logStatus = Logging;
    fileCreated = true;
  }
}

void task_called_after_GNSS_update(void) {
  //Get GNSS information
  GnssConf.getMessageInterval(&nmeaInterval.gga, &nmeaInterval.gll, &nmeaInterval.gsa, &nmeaInterval.gsv, &nmeaInterval.rmc, &nmeaInterval.vtg, &nmeaInterval.zda);
  GnssInfo.update();

  String logContent;    //NMEA sentence to store.
  int length =  GenerateNMEA(logContent);
  
  //Only log NMEA in fix mode, you can modify here for different logging mode
  if(GnssInfo.fixMode() < 2)
  {
    logStatus = NotFixed;
    return;
  }
  
  if(logStatus == NotFixed)
  {
    logStatus = Fixed;
  }
  
  //Always try to create log file when log status is fix mode. If successfully created, fix status will become logging.  
  if(logStatus == Fixed && !(CONTINUOUS_WRITE && fileCreated))  //Create file
  {
    CreateLogFile();
  }

  //No NMEA data need to store.
  if(length == 0)
    return;

  //Write NMEA sentence to file.
  File file = SD.open(logFile.c_str(), FILE_WRITE);
  if(file)
  {
    size_t s = file.write(logContent.c_str(), length);
    file.close();
    
    Serial.print("Write bytes : ");
    Serial.println(s);
    logStatus = Logging;
  }
  else
  {
    Serial.println("Write fail!");
    logStatus = Fixed;
  }
}
